import java.io.*;

//This is only a wrapper for the actual GRN, but it 
//provides some extra functionality for net analysis
public class GRNconverter extends KnabeGRN{

private input in;
public double lastIn;

public GRNconverter(int[] genome){
super(genome);
}

//will get the input currently set in KnabeGRN
//( set there via KnabeGRN.setInputFunction(whatFunction); )
public void getNewInput(){
this.in=getIn();
lastIn=.0;
}

//
public void getNewInput(int whatFunction){
KnabeGRN.setInputFunction(whatFunction);
this.in=getIn();
}

public boolean step(){
lastIn=in.getVal();
process(lastIn,0.0,0.0);
return true;
}

//returns a simplified network of the gene-gene interactions:
//M_{ij}=1 in the connectivity Matrix M if gene j encodes a protein
//that regulates gene i; M_{ij}=0 otherwise.
public network getNetwork(){
network ret=new network(this.genes.length);
ret.addNodes(this.genes.length);
KnabeGene ttt=null;
KnabeGene ttt2=null;
int[] curr;
for(int i=0;i<this.genes.length;i++){//go through all proteins and check their output
 ttt=(KnabeGene) this.genes[i];
 if(ttt==null) continue;
for(int h=0;h<this.genes.length;h++){//go through all genes to see who uses ttt's output
 ttt2=(KnabeGene) this.genes[h];
 if(ttt2==null) continue;
 for(int j=0;j<ttt2.geneCisReceptorSites.length;j++){
  curr=ttt2.getCisSite(j);
  //if(curr[0]==0) System.out.print("inhibitory");
  //else System.out.print("activatory");
  for(int l=1;l<curr.length;l++){//leaves 0 out because that has inhib./act. meaning
    if(curr[l]==ttt.geneOutputProtein){
     if(curr[0]==0)
      ret.link(i,h,true); //inhibitory
     else
      ret.link(i,h);
     }
    }
 }
}
}
return ret;
}

//returns a simplified network of the gene-gene interactions:
//M_{ij}=1 in the connectivity Matrix M if gene j encodes a protein
//that regulates gene i; M_{ij}=0 otherwise.
//HERE THE ENVIRONMENT IS INCLUDED AS A NODE WITH INPUT AND OUTPUT!
public network getNetworkWithEnvironment(){
network ret=new network(this.genes.length);
ret.addNodes(this.genes.length+1);
int envNode=this.genes.length;
KnabeGene ttt=null;
KnabeGene ttt2=null;
int[] curr;
for(int i=0;i<this.genes.length;i++){//go through all protein and check their output
 ttt=(KnabeGene) this.genes[i];
 if(ttt==null) continue;

if(ttt.geneOutputProtein==6)//outputs to environment
 ret.link(i,envNode);

for(int h=0;h<this.genes.length;h++){//go through all genes to see who uses ttt's output
 ttt2=(KnabeGene) this.genes[h];
 if(ttt2==null) continue;
 for(int j=0;j<ttt2.geneCisReceptorSites.length;j++){
  curr=ttt2.getCisSite(j);
  //if(curr[0]==0) System.out.print("inhibitory");
  //else System.out.print("activatory");
  for(int l=1;l<curr.length;l++){//leaves 0 out because that has inhib./act. meaning
    if(curr[l]==ttt.geneOutputProtein)
      ret.link(i,h);

    if(curr[l]==1)//will be called lots of times but doesn't matter as we don't count doubles
      ret.link(envNode,h);//_h_ is influenced by the environment

    }
 }
}
}
return ret;
}

//returns a simplified network of the gene-gene interactions:
//M_{ij}=1 in the connectivity Matrix M if gene j encodes a protein
//that regulates gene i; M_{ij}=0 otherwise.
//HERE THE ENVIRONMENT IS INCLUDED AS A NODE WITH INPUT AND OUTPUT!
//ALSO THE INPUT THAT INDICATES THE CELL TYPE IS CONSIDERED AS ENVIRONMENTAL INPUT
public network getNetworkWithEnvironmentAndDifferentiation(){
network ret=new network(this.genes.length);
ret.addNodes(this.genes.length+1);
int envNode=this.genes.length;
KnabeGene ttt=null;
KnabeGene ttt2=null;
int[] curr;
for(int i=0;i<this.genes.length;i++){//go through all protein and check their output
 ttt=(KnabeGene) this.genes[i];
 if(ttt==null) continue;

if(ttt.geneOutputProtein==6)//outputs to environment
 ret.link(i,envNode);

for(int h=0;h<this.genes.length;h++){//go through all genes to see who uses ttt's output
 ttt2=(KnabeGene) this.genes[h];
 if(ttt2==null) continue;
 for(int j=0;j<ttt2.geneCisReceptorSites.length;j++){
  curr=ttt2.getCisSite(j);
  //if(curr[0]==0) System.out.print("inhibitory");
  //else System.out.print("activatory");
  for(int l=1;l<curr.length;l++){//leaves 0 out because that has inhib./act. meaning
    if(curr[l]==ttt.geneOutputProtein)
      ret.link(i,h);

    if(curr[l]==1)//will be called lots of times but doesn't matter as we don't count doubles
      ret.link(envNode,h);//_h_ is influenced by the environment
    if(curr[l]==3)//will be called lots of times but doesn't matter as we don't count doubles
      ret.link(envNode,h);//_h_ is influenced by the differentiation indicator

    }
 }
}
}
return ret;
}

//prints some structural information
public void showStructure(){
System.out.println("Be careful with drawing conclusions from the follwing structure data as unused as well as disfunctional stuff is hard to remove automatically and thus it is in!");
int [] protInteractions=new int[8];
double cisSitesNum=0;
int maxCisNum=0;
double bindingNum=0;
int maxBindNum=0;
int [] curr;
KnabeGene ttt=null;
for(int i=0;i<this.genes.length;i++){
 ttt=(KnabeGene) this.genes[i];
 if(ttt==null){
   System.out.println("Gene "+i+" is not used.");
   continue;
 }
 System.out.print("Gene "+i+" has "+ttt.geneCisReceptorSites.length+" cis-sites and is ");
 if(ttt.geneOutputActivity==0)
  System.out.print("default 'off'");
 else
  System.out.print("default 'on'");
 System.out.println(" and produces protein "+ttt.geneOutputProtein);
 cisSitesNum+=ttt.geneCisReceptorSites.length;
 if(maxCisNum<ttt.geneCisReceptorSites.length)
   maxCisNum=ttt.geneCisReceptorSites.length;
 for(int j=0;j<ttt.geneCisReceptorSites.length;j++){
  curr=ttt.getCisSite(j);
  System.out.print(" cis-site "+j+" is ");
  if(curr[0]==0)
    System.out.print("inhibitory");
  else
    System.out.print("activatory");
  System.out.print(" and needs proteins:");
  if(maxBindNum<curr.length-1)
    maxBindNum=curr.length-1;
  bindingNum+=curr.length-1;
  for(int l=1;l<curr.length-1;l++){
    protInteractions[curr[l]]++;
    System.out.print(curr[l]+"+");
    }
  System.out.println(""+curr[curr.length-1]);
  protInteractions[curr[curr.length-1]]++;
 }
}
int maxActiveProt=0;
double usedProteins=0;
double avgActiveProt=0;
for(int i=0;i<protInteractions.length;i++){
  if(protInteractions[i]!=0)
    usedProteins++;
  avgActiveProt+=protInteractions[i];
  if(maxActiveProt<protInteractions[i])
    maxActiveProt=protInteractions[i];
  System.out.println("Protein "+i+" is needed on "+protInteractions[i]+" cis-sites.");
  }
System.out.println("average number of cis-sites per gene (considering disfunctional genes!): "+(cisSitesNum/(this.genes.length*1.0))+" ; max: "+maxCisNum);
System.out.println("average number of binding sites per cis-site (again disfunctional included): "+(bindingNum/cisSitesNum)+" ; max: "+maxBindNum);
System.out.println("average number (over "+(int)usedProteins+" of 8 proteins used) of interactions a protein is involved: "+(avgActiveProt/usedProteins)+" ; max: "+maxActiveProt);
}

public static int[] stringToIntArr(String in){
int[] ret=new int[in.length()];
int j=0;
for(int i=0;i<ret.length;i++)
 try{
  ret[j]=Integer.parseInt(""+in.charAt(i));
  j++;
 }catch(Exception e){}
int[] ret2=new int[j];
System.arraycopy(ret,0,ret2,0,j);
return ret2;
}

public static GRNconverter GRNfromString(String genome){
String[] t=genome.split("3");
KnabeGRN.geneNo=t.length-1;
String gg=genome.replace(',',' ');
if(!gg.startsWith("{"))
  return null;
gg=gg.replace('{',' ');
gg=gg.replace('}',' ');
int[] geneInts=stringToIntArr(gg);
return new GRNconverter(geneInts);
}

public static GRNconverter readGRNfromFile(String filename){
String read =readGenomeFromFile(filename);
return GRNfromString(read);
}

public static String readGenomeFromFile(File filename){
BufferedReader geno =null;
String read ="";
try{
              geno=new BufferedReader(
                    new FileReader(
                    filename));
read= geno.readLine(); //fitness data only
read= geno.readLine(); //the genome
return read;
}
catch(Exception e){System.err.println("FILE ERROR in readGenomeFromFile() "+e);}
return null;
}

public static String readGenomeFromFile(String filename){
BufferedReader geno =null;
String read ="";
try{
              geno=new BufferedReader(
                    new FileReader(
                    filename));
read= geno.readLine(); //fitness data only
read= geno.readLine(); //the genome
return read;
}
catch(Exception e){System.err.println("FILE ERROR in readGenomeFromFile() "+e);}
return null;
}

} 
